
from bokchoi.config import Config
from bokchoi.aws import EMR, EC2
from bokchoi.gcp import GCP


def requires_config(fn):
    """
    Decorator; adds check whether config was loaded successfully.
    :param fn:                      Function to add check to
    :return:                        Decorated function
    """
    def fn_check_for_config(self, *args, **kwargs):
        if not self.config.loaded:
            return fn.__name__ + ' requires config. Run \'bokchoi init NAME\' to initialise.'
        return fn(self, *args, **kwargs)
    return fn_check_for_config


class Bokchoi:

    backends = {'EC2': EC2, 'EMR': EMR, 'GCP': GCP}

    def __init__(self, path):

        self.config = Config(path)

        try:
            self.config.load()
        except FileNotFoundError:
            print('Config not found')
        else:
            self.backend = self.backends[self.config['Platform']](self.config.name, self.config)

    def init(self, name, platform):
        """ Initialise new project
        :param name:                Name of the project
        :param platform:            Platform used to run application
        :return:                    Response object
        """

        if self.config.loaded:
            return 'Project already initialised. Deploy using \'bokchoi deploy\'.'
        self.config.init(name, platform, self.backends[platform].default_config)

        return 'Project initialised. Deploy using \'bokchoi deploy\'.'

    @requires_config
    def deploy(self):
        print('Deploying: ' + self.config.name)
        return self.backend.deploy(path=self.config.path)

    @requires_config
    def undeploy(self, dryrun):
        print('Undeploying: ' + self.config.name)
        return self.backend.undeploy(dryrun)

    @requires_config
    def run(self):
        print('Running: ' + self.config.name)
        return self.backend.run()

    @requires_config
    def stop(self, *args, **kwargs):
        return self.backend.stop(*args, **kwargs)

    @requires_config
    def connect(self, dryrun, *args, **kwargs):
        self.backend.connect(dryrun, *args, **kwargs)

    @requires_config
    def status(self):
        return self.backend.status()

    @requires_config
    def logs(self):
        return self.backend.logs()



"""
Module contains functions related to SSH.

Most of this script is adapted from:
https://github.com/paramiko/paramiko/blob/master/demos/forward.py
"""
import os
import select
import socketserver

from paramiko import RSAKey, SSHClient, AutoAddPolicy
from paramiko.ssh_exception import SSHException

from bokchoi import utils


class ForwardServer(socketserver.ThreadingTCPServer):
    daemon_threads = True
    allow_reuse_address = True


class Handler(socketserver.BaseRequestHandler):

    ssh_transport = None
    host_port = None
    remote_port = None

    def __init__(self, request, client_address, server):
        super(Handler, self).__init__(request, client_address, server)

    def handle(self):

        channel = self.ssh_transport.open_channel('direct-tcpip',
                                                  ('localhost', self.remote_port),
                                                  ('localhost', self.host_port))
        if not channel:
            print('Incoming request was rejected')
            return

        while True:
            r, _, _ = select.select([self.request, channel], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if not data:
                    break
                channel.send(data)
            if channel in r:
                data = channel.recv(1024)
                if not data:
                    break
                self.request.send(data)

        channel.close()
        self.request.close()


class SSH(object):

    def __init__(self, private_key_name):
        self.client = SSHClient()
        self.client.load_system_host_keys()
        self.client.set_missing_host_key_policy(AutoAddPolicy())

        self.public_key, self.key_file_path = self._maybe_generate_keys(private_key_name)

    def forward(self, local_port, remote_host, remote_po, user_name):
        """ Sets up port forwarding to remote host
        :param local_port:              Local port to forward to
        :param remote_host:             Remote host
        :param remote_port:             Remote port
        :param user_name:               User to use in ssh connection
        :return:                        -
        """

        print('Connecting to ssh host {}:{} ...'.format(remote_host, remote_po))

        utils.retry(self.client.connect
                    , SSHException
                    , hostname=remote_host
                    , port=22
                    , username=user_name
                    , key_filename=self.key_file_path)

        class SubHandler(Handler):
            ssh_transport = self.client.get_transport()
            host_port = local_port
            remote_port = remote_po

        server = ForwardServer(('localhost', local_port), SubHandler)

        try:
            server.serve_forever()
        except KeyboardInterrupt:
            server.shutdown()
            server.server_close()
            print('Connection closed')

    def _maybe_generate_keys(self, private_key_name):
        """Get private and public keys. Create if not exists.
        :return:                    Public key
        """
        ssh_dir = os.path.join(os.path.expanduser('~'), '.ssh')
        if not os.path.exists(ssh_dir):
            os.makedirs(ssh_dir)

        key_file_path = os.path.join(ssh_dir, private_key_name)
        try:
            priv = RSAKey.from_private_key_file(key_file_path)
        except FileNotFoundError:
            print('Could not find private key. Creating one.')
            priv = RSAKey.generate(2048)
            priv.write_private_key_file(key_file_path)
        pub = priv.get_base64()
        return pub, key_file_path




import hashlib
from time import sleep
import urllib
from io import BytesIO
import os
import zipfile

from bokchoi.aws import cloudwatch_logger


def retry(func, exc, **kwargs):
    """ Retries boto3 function call in case a ClientError occurs
    :param func:                    Function to call
    :param exc:                     Exception to catch
    :param kwargs:                  Parameters to pass to function
    :return:                        Function response
    """
    for _ in range(60):
        try:
            response = func(**kwargs)
            return response
        except exc:
            sleep(1)

    raise TimeoutError()


def create_project_id(project_name, vendor_specific_id):
    """Creates project id by hashing vendor specific id and project name"""
    unique_id = hashlib.sha1((vendor_specific_id + project_name).encode()).hexdigest()
    return '-'.join(('bokchoi', project_name, unique_id[:12]))


def get_my_ip():
    with urllib.request.urlopen('https://api.ipify.org/') as response:
        return response.read().decode('utf8')


def zip_package(path, requirements=None):
    """ Creates deployment package by zipping the project directory. Writes requirements to requirements.txt
    if specified in settings
    :param path:                    Path to project directory
    :param requirements:            List of python requirements
    :return:                        Zip file
    """
    file_object = BytesIO()

    rootlen = len(path) + 1

    with zipfile.ZipFile(file_object, 'w', zipfile.ZIP_DEFLATED) as zip_file:
        for base, _, files in os.walk(path):
            for file_name in files:
                fn = os.path.join(base, file_name)
                zip_file.write(fn, fn[rootlen:])

        zip_file.write(cloudwatch_logger.__file__, 'cloudwatch_logger.py')

        zip_file.writestr('requirements.txt', '\n'.join(requirements or ''))

        fingerprint = '|'.join([str(elem.CRC) for elem in zip_file.infolist()])

    file_object.seek(0)

    return file_object, fingerprint



#!/usr/bin/env python3

import os
import sys
import time

import boto3


class CloudwatchLogger:

    """Reads messages from stdin and logs them to Cloudwatch Logs"""

    def __init__(self):

        self.logs_client = boto3.client('logs', region_name=os.environ['REGION'])

        self.log_group_name = os.environ['BOKCHOI_PROJECT_ID']
        self.log_stream_name, self.sequence_token = self.get_most_recent_log_stream(self.log_group_name)

        self.stage = sys.argv[1]

    def get_most_recent_log_stream(self, log_group_name):
        """Returns most recent log stream. Should always exist since it's
        created when 'bokchoi run' is executed"""
        response = self.logs_client.describe_log_streams(
            logGroupName=log_group_name,
            orderBy='LogStreamName',
            descending=True,
            limit=1
        )

        log_stream = response['logStreams'][0]
        next_token = log_stream.get('uploadSequenceToken')

        return log_stream['logStreamName'], next_token

    def log_message(self, message):
        """Log message to Cloudwatch log group"""
        log_info = {'logGroupName': self.log_group_name,
                    'logStreamName': self.log_stream_name,
                    'logEvents': [
                        {
                            'timestamp': int(1000 * time.time()),
                            'message': '[{}]: {}'.format(self.stage, message)
                        },
                    ]}

        if self.sequence_token:
            log_info['sequenceToken'] = self.sequence_token

        response = self.logs_client.put_log_events(**log_info)

        self.sequence_token = response['nextSequenceToken']

    def run(self):
        """Process incoming messages"""

        for message in sys.stdin:
            self.log_message(message)


if __name__ == '__main__':
    CloudwatchLogger().run()
